<?php
	include 'compiler/Node.php';
	
	final class Blueprint
	{
		const E_OK = 0;
		const E_OPENFAILED = -1;
		const E_UNKNOWNNODE = -2;
		const E_INVALIDFORMAT = -3;
		
		const COMPILER_IDLE = 0;
		const COMPILER_PROCESS = 1;
		const COMPILER_ERROR = 2;
		const COMPILER_DONE = 3;
		
		const MOVE_UP = 0;
		const MOVE_DOWN = 1;
		
		const DIR_FORWARD = 0;
		const DIR_BACKWARD = 1;
		
		private $includes;
		private $main;
		private $nodes;
		private $name;
		
		// Compiler state
		private $eIndex;
		private $eMessage;
		private $currentState;
		private $tabCount;
		
		private $output;
		
		public function __construct()
		{
			$this->clear();
		}
		
		public function clear()
		{
			$this->includes = array();
			$this->nodes = array();
			$this->main = false;
			$this->currentState = self::COMPILER_IDLE;
			$this->error = null;
			$this->output = "";
			$this->name = "";
		}
		
		public function isMain()
		{
			return $this->main;
		}
		
		public function setMain($main)
		{
			$this->main = $main;
		}
		
		public function setName($name)
		{
			if(is_string($name) && strlen($name) > 0)
				$this->name = $name;
		}
		
		public function load($fileName)
		{
			if(file_exists($fileName))
			{
				$f = file_get_contents($fileName);
				
				if(!$f)
					return self::E_OPENFAILED;
				
				$json = json_decode($f, true);
				
				if(isset($json["includes"]) && is_array($json["includes"]))
				{
					foreach($json["includes"] as $include)
					{
						if(is_string($include) && file_exists($include))
							$this->includes[] = $include;
					}
				}
				
				if(!isset($json["bp"]))
					return self::E_INVALIDFORMAT;
				
				if(isset($json["nodes"]) && is_array($json["nodes"]))
				{
					foreach($json["nodes"] as $node)
					{
						if(isset($node["nname"]))
						{
							$instance = $this->createNode($node["nname"]);
							
							if(!$instance)
								return self::E_UNKINSTANCE;
							
							$attrs = $instance->getFactory()->queryAttributeList();
							
							foreach($attrs as $attr)
							{
								$instance->setAttribute($attr["name"], $node[$attr["name"]]);
							}
						}
					}
				}
				
				return self::E_OK;
			}
			
			return self::E_OPENFAILED;
		}
		
		public function save($fileName)
		{
			$json = array("bp" => true);
			
			foreach($this->includes as $include)
			{
				if(file_exists($include))
					$json["includes"][] = $include;
			}
			
			foreach($this->nodes as $node)
			{
				if(!$node)
					continue;
				
				$attrs = $node->getFactory()->queryAttributeList();
				$nstruct = array();
				
				$nstruct["nname"] = $node->getFactory()->getName();
				
				foreach($attrs as $attr)
				{
					$nstruct[$attr["name"]] = $node->getAttribute($attr["name"]);
				}
				
				$json["nodes"][] = $nstruct;
			}
			
			return file_put_contents($fileName, json_encode($json));
		}
		
		public function throwError($error)
		{	
			if($this->currentState == self::COMPILER_PROCESS)
			{
				$this->eMessage = "An error occured while compiling node #$this->eIndex: $error";
				
				$this->currentState = self::COMPILER_ERROR;
			}
		}
		
		public function getCurrentState()
		{
			return $this->currentState;
		}
		
		public function getErrorMessage()
		{
			return $this->eMessage;
		}
		
		public function getOutput()
		{
			return $this->output;
		}
		
		public function getName()
		{
			return $this->name;
		}
		
		public function compile()
		{
			if($this->currentState == self::COMPILER_PROCESS)
				return null;
			
			$start = time();
			
			$this->eIndex = 0;
			$this->tabCount = 0;
			$this->output = "";
			$this->currentState = self::COMPILER_PROCESS;
			
			foreach($this->includes as $inc)
			{
				$bp = new Blueprint();
				
				if($bp->load($inc))
					$this->output .= $bp->compile();
				
				$bp = null;
			}
			
			if(strlen($this->name) < 1)
			{
				$this->throwError("Blueprint should have a name");
				
				return false;
			}
			
			$this->output .= "<?php ";
			$this->output .= "function $this->name()";
			$this->output .= "{";
			
			foreach($this->nodes as $node)
			{
				if($this->currentState == self::COMPILER_ERROR)
					return false;
				
				if(get_class($node) == "StartBlockNode")
					$this->tabCount++;
				
				if(get_class($node) == "EndBlockNode")
					$this->tabCount--;
				
				$this->output .= $node->compile();
				$this->eIndex++;
			}
			
			$this->output .= "}";
			
			if($this->isMain())
				$this->output .= "$this->name();";
			
			if($this->tabCount != 0)
			{
				$this->throwError("Invalid block start/end pair");
				
				return false;
			}
			
			$this->output .= " ?>";
			
			$end = $start - time();
			
			$this->eMessage = "Compilation successful in $end s. (compiled $this->eIndex nodes)";
			$this->currentState = Blueprint::COMPILER_DONE;
			
			return true;
		}
		
		public function addInclude($include)
		{
			if(is_string($include) && strlen($include) > 0 && file_exists($include))
				$this->includes = $include;
		}
		
		public function removeInclude($include)
		{
			array_filter($this->includes, function($val) {
				if($val == $include)
					return false;
				
				return true;
			});
		}
		
		public function getIncludes()
		{
			return $this->includes;
		}
		
		public function getNodes()
		{
			return $this->nodes;
		}
		
		public function getNode($i)
		{
			return isset($this->nodes[$i]) ? $this->nodes[$i] : null;
		}
		
		private function arrayFilter($array, $callback, $data)
		{
			$ret = array();
			
			foreach($array as $k => $v)
			{
				if($callback($k, $v, $data))
					$ret[] = $v;
			}
			
			return $ret;
		}
		
		public function getNodeCount()
		{
			return sizeof($this->nodes);
		}
		
		public function indexOf($node)
		{
			for($i = 0; $i < sizeof($this->nodes); $i++)
			{
				if($this->nodes[$i] === $node)
					return $i;
			}
			
			return -1;
		}
		
		public function removeNode($node)
		{
			$this->nodes = $this->arrayFilter($this->nodes, function($key, $value, $data)
			{
				if($value === $data)
					return false;
				
				return true;
			}, $node);
		}
		
		public function moveNode($node, $dir)
		{
			$index = $this->indexOf($node);
			
			if($index >= 0)
			{
				if($dir == Blueprint::MOVE_DOWN)
				{
					if($index < sizeof($this->nodes) - 1)
					{
						$a = $this->nodes[$index + 1];
						$this->nodes[$index] = $a;
						$this->nodes[$index + 1] = $node;
					}
				}
				else
				{
					if($index > 0)
					{
						$a = $this->nodes[$index - 1];
						$this->nodes[$index] = $a;
						$this->nodes[$index - 1] = $node;
					}
				}
			}
		}
		
		public function createNode($name)
		{
			$factory = NodeList::findFactory($name);
			
			if($factory)
			{
				$node = $factory->instantiate($this);
				
				if($node)
					$this->nodes[] = $node;
				
				return $node;
			}
			
			return null;
		}
	}